@@ -2,6 +2,8 @@ from apache_beam.utils.timestamp import MIN_TIMESTAMP
 from apache_beam.utils.windowed_value import WindowedValue
 from apache_beam.transforms.window import GlobalWindow
 from apache_beam.transforms import PTransform, ParDo, DoFn
+from scipy.sparse import hstack, vstack, csr_matrix
+import numpy as np
 from feature_creation_tools.feature_creation_utils import LabelEncoderExt
 import google.cloud.storage as storage
 import pandas as pd
@@ -460,12 +462,249 @@ class CreateCategoricalEncodedVariables(PTransform):
 class _SaveDataframeToGCS(DoFn):
 
     def process(self, df):
-        bundle_num = df.iloc[0]["person_id"]
+
+        person_id_column = df[['person_id']].copy()
+        bundle_num = person_id_column.iloc[0]["person_id"]
+
+        # drop person_id column
+        df.drop('person_id', axis=1, inplace=True)
+
+        # Drop useless features for now
+        df.drop('safer_result_oosr_hazmat_us', axis=1, inplace=True)
+        df.drop('safer_result_oosr_vehicle_ca', axis=1, inplace=True)
+        df.drop('safer_result_oosr_vehicle_us', axis=1, inplace=True)
+        df.drop('operating_lanes_1_pickup_state_encoded', axis=1, inplace=True)
+        df.drop('operating_lanes_2_pickup_state_encoded', axis=1, inplace=True)
+        df.drop('operating_lanes_3_pickup_state_encoded', axis=1, inplace=True)
+        df.drop('operating_lanes_4_pickup_state_encoded', axis=1, inplace=True)
+        df.drop('operating_lanes_5_pickup_state_encoded', axis=1, inplace=True)
+        df.drop('operating_lanes_1_dropoff_state_encoded', axis=1, inplace=True)
+        df.drop('operating_lanes_2_dropoff_state_encoded', axis=1, inplace=True)
+        df.drop('operating_lanes_3_dropoff_state_encoded', axis=1, inplace=True)
+        df.drop('operating_lanes_4_dropoff_state_encoded', axis=1, inplace=True)
+        df.drop('operating_lanes_5_dropoff_state_encoded', axis=1, inplace=True)
+        df.drop('operating_lanes_1_pickup_city_encoded', axis=1, inplace=True)
+        df.drop('operating_lanes_2_pickup_city_encoded', axis=1, inplace=True)
+        df.drop('operating_lanes_3_pickup_city_encoded', axis=1, inplace=True)
+        df.drop('operating_lanes_4_pickup_city_encoded', axis=1, inplace=True)
+        df.drop('operating_lanes_5_pickup_city_encoded', axis=1, inplace=True)
+        df.drop('operating_lanes_1_dropoff_city_encoded', axis=1, inplace=True)
+        df.drop('operating_lanes_2_dropoff_city_encoded', axis=1, inplace=True)
+        df.drop('operating_lanes_3_dropoff_city_encoded', axis=1, inplace=True)
+        df.drop('operating_lanes_4_dropoff_city_encoded', axis=1, inplace=True)
+        df.drop('operating_lanes_5_dropoff_city_encoded', axis=1, inplace=True)
+        df.fillna(0, inplace=True)
+
+        # fix bool values
+        df['safer_result_allowed_hire'] = (df['safer_result_allowed_hire'] == 'True').astype(int)
+        df['safer_result_cargo_carried_agricultural_farm_supplies'] = (
+                df['safer_result_cargo_carried_agricultural_farm_supplies'] == 'True').astype(int)
+        df['safer_result_cargo_carried_beverages'] = (
+                df['safer_result_cargo_carried_beverages'] == 'True').astype(int)
+        df['safer_result_cargo_carried_building_materials'] = (
+                df['safer_result_cargo_carried_building_materials'] == 'True').astype(int)
+        df['safer_result_cargo_carried_chemicals'] = (
+                df['safer_result_cargo_carried_chemicals'] == 'True').astype(int)
+        df['safer_result_cargo_carried_coal_coke'] = (
+                df['safer_result_cargo_carried_coal_coke'] == 'True').astype(int)
+        df['safer_result_cargo_carried_commodities_dry_bulk'] = (
+                df['safer_result_cargo_carried_commodities_dry_bulk'] == 'True').astype(int)
+        df['safer_result_cargo_carried_construction'] = (
+                df['safer_result_cargo_carried_construction'] == 'True').astype(int)
+        df['safer_result_cargo_carried_drive_tow_away'] = (
+                df['safer_result_cargo_carried_drive_tow_away'] == 'True').astype(int)
+        df['safer_result_cargo_carried_fresh_produce'] = (
+                df['safer_result_cargo_carried_fresh_produce'] == 'True').astype(int)
+        df['safer_result_cargo_carried_garbage_refuse'] = (
+                df['safer_result_cargo_carried_garbage_refuse'] == 'True').astype(int)
+        df['safer_result_cargo_carried_general_freight'] = (
+                df['safer_result_cargo_carried_general_freight'] == 'True').astype(int)
+        df['safer_result_cargo_carried_grain_feed_hay'] = (
+                df['safer_result_cargo_carried_grain_feed_hay'] == 'True').astype(int)
+        df['safer_result_cargo_carried_household_goods'] = (
+                df['safer_result_cargo_carried_household_goods'] == 'True').astype(int)
+        df['safer_result_cargo_carried_intermodal_cont'] = (
+                df['safer_result_cargo_carried_intermodal_cont'] == 'True').astype(int)
+        df['safer_result_cargo_carried_liquids_gases'] = (
+                df['safer_result_cargo_carried_liquids_gases'] == 'True').astype(int)
+        df['safer_result_cargo_carried_livestock'] = (
+                df['safer_result_cargo_carried_livestock'] == 'True').astype(int)
+        df['safer_result_cargo_carried_logs_poles_beams_lumber'] = (
+                df['safer_result_cargo_carried_logs_poles_beams_lumber'] == 'True').astype(int)
+        df['safer_result_cargo_carried_machinery_large_objects'] = (
+                df['safer_result_cargo_carried_machinery_large_objects'] == 'True').astype(int)
+        df['safer_result_cargo_carried_meat'] = (df['safer_result_cargo_carried_meat'] == 'True').astype(int)
+        df['safer_result_cargo_carried_metal_sheets_coils_rolls'] = (
+                df['safer_result_cargo_carried_metal_sheets_coils_rolls'] == 'True').astype(int)
+        df['safer_result_cargo_carried_mobile_homes'] = (
+                df['safer_result_cargo_carried_mobile_homes'] == 'True').astype(int)
+        df['safer_result_cargo_carried_motor_vehicles'] = (
+                df['safer_result_cargo_carried_motor_vehicles'] == 'True').astype(int)
+        df['safer_result_cargo_carried_oilfield_equipment'] = (
+                df['safer_result_cargo_carried_oilfield_equipment'] == 'True').astype(int)
+        df['safer_result_cargo_carried_paper_products'] = (
+                df['safer_result_cargo_carried_paper_products'] == 'True').astype(int)
+        df['safer_result_cargo_carried_passengers'] = (
+                df['safer_result_cargo_carried_passengers'] == 'True').astype(int)
+        df['safer_result_cargo_carried_refrigerated_food'] = (
+                df['safer_result_cargo_carried_refrigerated_food'] == 'True').astype(int)
+        df['safer_result_cargo_carried_us_mail'] = (
+                df['safer_result_cargo_carried_us_mail'] == 'True').astype(
+            int)
+        df['safer_result_cargo_carried_utilities'] = (
+                df['safer_result_cargo_carried_utilities'] == 'True').astype(int)
+        df['safer_result_cargo_carried_water_well'] = (
+                df['safer_result_cargo_carried_water_well'] == 'True').astype(int)
+
+        # expand lists in dataframe
+        split_out = df['equipment_types_encoded'].apply(pd.Series)
+        split_out = split_out.rename(columns=lambda x: 'equipment_types_encoded_' + str(x))
+        df = pd.concat([df, split_out], axis=1, sort=False)
+        df.drop('equipment_types_encoded', axis=1, inplace=True)
+
+        split_out = df['predicted_equipment_types_encoded'].apply(pd.Series)
+        split_out = split_out.rename(columns=lambda x: 'predicted_equipment_types_encoded_' + str(x))
+        df = pd.concat([df, split_out], axis=1, sort=False)
+        df.drop('predicted_equipment_types_encoded', axis=1, inplace=True)
+
+        # Combine all the sparse column features into a csr matrix
+        sparse_users = vstack(df['state_encoded'].values)
+        df.drop('state_encoded', axis=1, inplace=True)
+
+        city = vstack(df['city_encoded'].values)
+        df.drop('city_encoded', axis=1, inplace=True)
+        sparse_users = hstack((sparse_users, city))
+
+        state = vstack(df['common_lane_1_pickup_state_encoded'].values)
+        df.drop('common_lane_1_pickup_state_encoded', axis=1, inplace=True)
+        sparse_users = hstack((sparse_users, state))
+
+        state = vstack(df['common_lane_2_pickup_state_encoded'].values)
+        df.drop('common_lane_2_pickup_state_encoded', axis=1, inplace=True)
+        sparse_users = hstack((sparse_users, state))
+
+        state = vstack(df['common_lane_3_pickup_state_encoded'].values)
+        df.drop('common_lane_3_pickup_state_encoded', axis=1, inplace=True)
+        sparse_users = hstack((sparse_users, state))
+
+        state = vstack(df['common_lane_4_pickup_state_encoded'].values)
+        df.drop('common_lane_4_pickup_state_encoded', axis=1, inplace=True)
+        sparse_users = hstack((sparse_users, state))
+
+        state = vstack(df['common_lane_5_pickup_state_encoded'].values)
+        df.drop('common_lane_5_pickup_state_encoded', axis=1, inplace=True)
+        sparse_users = hstack((sparse_users, state))
+
+        state = vstack(df['common_lane_1_dropoff_state_encoded'].values)
+        df.drop('common_lane_1_dropoff_state_encoded', axis=1, inplace=True)
+        sparse_users = hstack((sparse_users, state))
+
+        state = vstack(df['common_lane_2_dropoff_state_encoded'].values)
+        df.drop('common_lane_2_dropoff_state_encoded', axis=1, inplace=True)
+        sparse_users = hstack((sparse_users, state))
+
+        state = vstack(df['common_lane_3_dropoff_state_encoded'].values)
+        df.drop('common_lane_3_dropoff_state_encoded', axis=1, inplace=True)
+        sparse_users = hstack((sparse_users, state))
+
+        state = vstack(df['common_lane_4_dropoff_state_encoded'].values)
+        df.drop('common_lane_4_dropoff_state_encoded', axis=1, inplace=True)
+        sparse_users = hstack((sparse_users, state))
+
+        state = vstack(df['common_lane_5_dropoff_state_encoded'].values)
+        df.drop('common_lane_5_dropoff_state_encoded', axis=1, inplace=True)
+        sparse_users = hstack((sparse_users, state))
+
+        city = vstack(df['common_lane_1_pickup_city_encoded'].values)
+        df.drop('common_lane_1_pickup_city_encoded', axis=1, inplace=True)
+        sparse_users = hstack((sparse_users, city))
+
+        city = vstack(df['common_lane_2_pickup_city_encoded'].values)
+        df.drop('common_lane_2_pickup_city_encoded', axis=1, inplace=True)
+        sparse_users = hstack((sparse_users, city))
+
+        city = vstack(df['common_lane_3_pickup_city_encoded'].values)
+        df.drop('common_lane_3_pickup_city_encoded', axis=1, inplace=True)
+        sparse_users = hstack((sparse_users, city))
+
+        city = vstack(df['common_lane_4_pickup_city_encoded'].values)
+        df.drop('common_lane_4_pickup_city_encoded', axis=1, inplace=True)
+        sparse_users = hstack((sparse_users, city))
+
+        city = vstack(df['common_lane_5_pickup_city_encoded'].values)
+        df.drop('common_lane_5_pickup_city_encoded', axis=1, inplace=True)
+        sparse_users = hstack((sparse_users, city))
+
+        city = vstack(df['common_lane_1_dropoff_city_encoded'].values)
+        df.drop('common_lane_1_dropoff_city_encoded', axis=1, inplace=True)
+        sparse_users = hstack((sparse_users, city))
+
+        city = vstack(df['common_lane_2_dropoff_city_encoded'].values)
+        df.drop('common_lane_2_dropoff_city_encoded', axis=1, inplace=True)
+        sparse_users = hstack((sparse_users, city))
+
+        city = vstack(df['common_lane_3_dropoff_city_encoded'].values)
+        df.drop('common_lane_3_dropoff_city_encoded', axis=1, inplace=True)
+        sparse_users = hstack((sparse_users, city))
+
+        city = vstack(df['common_lane_4_dropoff_city_encoded'].values)
+        df.drop('common_lane_4_dropoff_city_encoded', axis=1, inplace=True)
+        sparse_users = hstack((sparse_users, city))
+
+        city = vstack(df['common_lane_5_dropoff_city_encoded'].values)
+        df.drop('common_lane_5_dropoff_city_encoded', axis=1, inplace=True)
+        sparse_users = hstack((sparse_users, city))
+
+        zip_code = vstack(df['zip_code_encoded'].values)
+        df.drop('zip_code_encoded', axis=1, inplace=True)
+        sparse_users = hstack((sparse_users, zip_code))
+
+        email = vstack(df['email_encoded'].values)
+        df.drop('email_encoded', axis=1, inplace=True)
+        sparse_users = hstack((sparse_users, email))
+
+        dot = vstack(df['dot_encoded'].values)
+        df.drop('dot_encoded', axis=1, inplace=True)
+        sparse_users = hstack((sparse_users, dot))
+
+        channel = vstack(df['channel_encoded'].values)
+        df.drop('channel_encoded', axis=1, inplace=True)
+        sparse_users = hstack((sparse_users, channel))
+
+        hos_cycle = vstack(df['hos_cycle_encoded'].values)
+        df.drop('hos_cycle_encoded', axis=1, inplace=True)
+        sparse_users = hstack((sparse_users, hos_cycle))
+
+        hos_home_timezone = vstack(df['hos_home_timezone_encoded'].values)
+        df.drop('hos_home_timezone_encoded', axis=1, inplace=True)
+        sparse_users = hstack((sparse_users, hos_home_timezone))
+
+        mc_number = vstack(df['mc_number_encoded'].values)
+        df.drop('mc_number_encoded', axis=1, inplace=True)
+        sparse_users = hstack((sparse_users, mc_number))
+
+        user_type = vstack(df['user_type_encoded'].values)
+        df.drop('user_type_encoded', axis=1, inplace=True)
+        sparse_users = hstack((sparse_users, user_type))
+
+        safer_result_safety_rating = vstack(df['safer_result_safety_rating_encoded'].values)
+        df.drop('safer_result_safety_rating_encoded', axis=1, inplace=True)
+        sparse_users = hstack((sparse_users, safer_result_safety_rating))
+
+        # combine remaining dataframe features into a csr matrix
+        remaining_features = csr_matrix(df.values)
+        sparse_users = hstack((sparse_users, remaining_features))
+
+        # Add Identity Matrix to feature matrix as suggested in documentation
+        sparse_users.data = np.nan_to_num(sparse_users.data, copy=False)
+        sparse_users = csr_matrix(sparse_users)
+
+        data = {"index": person_id_column, "data": sparse_users}
         filename = "user_features_" + str(bundle_num) + ".pkl"
-        df.to_pickle(filename)
+        pickle.dump(data, open(filename, "wb"))
         upload_blob("user_features", filename, filename)
 
-        return [df, ]
+        return [data, ]
 
 
 class SaveDataframeToGCS(PTransform):
